from .utils import contribute_to_dict


class TableColumn:
    """Represents a table column."""

    def __init__(self, name, title=None, source=None, sort=None, search=None, default=None, css=None):
        """
        :param str name: Column name (alias).
        :param str title: Column title.
        :param source: Column source object
        :param bool sort: Flag to toggle sorting.
        :param bool search: Flag to toggle searching.
        :param str default: Value to be used if no data available.
        :param str css: CSS class for the column.

        """
        self.name = name
        self.title = title or name
        self.source = source
        self.sort = sort
        self.search = search
        self.default = default
        self.css = css

        # todo decide on `render` js

    def as_dict(self):
        """Represents the column as a dictionary.

        :rtype: dict
        """
        name = self.name

        result = {
            'data': name,
            'name': name,
            'title': '%s' % self.title,
        }

        contribute_to_dict([
            ('sortable', self.sort),
            ('searchable', self.search),
            ('defaultContent', self.default),
            ('className', self.css),
        ], to=result)

        return result



from collections import OrderedDict
from itertools import chain

from django.db.models import QuerySet

from .columns import TableColumn


class TableSource:
    """Base data source for tables."""

    def __init__(self, source, options=None):
        self.columns = OrderedDict()
        self.row_id = 'DT_RowId'
        self.options = options or {}
        self._rows = []
        self._bootstrap(source)

    @classmethod
    def spawn(cls, source, params):
        """Alternative constructor.

        :param source:
        :param dict params:
        :rtype: TableSource

        """
        source_obj = cls(source, options=params.get('options'))
        return source_obj

    def _bootstrap(self, source):
        """The place for a source-specific bootstrap."""

    def contribute_to_config(self, config, table):
        """Updates table configuration dictionary with source-specific params.

        :param dict config:
        :param Table table:

        """
        config.update({
            'createdRow': lambda: (
                "function(row, data, idx){var v=data['%s']; if (v){$(row).attr('data-id', v);}}" % self.row_id),

            'processing': True,
            'columns': [column.as_dict() for column in self.columns.values()],
        })

        options = self.options

        if not options.get('init_dom'):
            # todo maybe use serialization instead of string casting
            # todo FK support
            config['data'] = [{k: '%s' % v for k, v in row.items()} for row in self.rows]

    @property
    def rows(self):
        """Represents table rows.

        :rtype: list[dict]
        """
        return self._rows


class ListDictsSource(TableSource):
    """Static data source.

    .. code-block:: python

        source = [
            {
                'one': '1',
                'two': '2',
            },
            {
                'one': '3',
                'two': '4',
            },
        ]

    """

    def _bootstrap(self, source):
        super()._bootstrap(source)

        columns = self.columns

        names = list(source[0].keys())

        for name in names:
            columns[name] = TableColumn(name=name)

        self._rows = source
        self.row_id = names[0]  # Use first column value.


class ModelSource(TableSource):
    """Django model datasource.

    .. code-block:: python

        source = Article  # Model class.
        source = Article.objects.filter(hidden=False)  # Or a QuerySet.

    """

    def _bootstrap(self, source):
        super()._bootstrap(source)

        columns = self.columns

        if isinstance(source, QuerySet):
            model = source.model
            qs = source

        else:  # Model class
            model = source
            qs = model.objects.all()

        opts = model._meta

        for field in chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
            name = field.name
            columns[name] = TableColumn(name=name, title=field.verbose_name, source=field)

        self.model = model
        self.qs = qs

        self.row_id = opts.pk.name

    @property
    def rows(self):
        columns = self.columns
        result = self.qs.values(*columns.keys())
        return result



from django.utils.translation import get_language

from .base import TablePlugin


LANGS = {
    'af': 'Afrikaans',
    'sq': 'Albanian',
    'am': 'Amharic',
    'ar': 'Arabic',
    'hy': 'Armenian',
    'az': 'Azerbaijan',
    'bn': 'Bangla',
    'eu': 'Basque',
    'be': 'Belarusian',
    'bg': 'Bulgarian',
    'ca': 'Catalan',
    'zh-hant': 'Chinese-traditional',
    'zh-hans': 'Chinese',
    'zh': 'Chinese',
    'hr': 'Croatian',
    'cs': 'Czech',
    'da': 'Danish',
    'nl': 'Dutch',
    # 'en': 'English',  # this on is built-in
    'et': 'Estonian',
    'fil': 'Filipino',  # no two-letter code
    'fi': 'Finnish',
    'fr': 'French',
    'gl': 'Galician',
    'ka': 'Georgian',
    'de': 'German',
    'el': 'Greek',
    'gu': 'Gujarati',
    'he': 'Hebrew',
    'hi': 'Hindi',
    'hu': 'Hungarian',
    'is': 'Icelandic',
    'ida': 'Indonesian-Alternative',  # pseudo code
    'id': 'Indonesian',
    'ga': 'Irish',
    'it': 'Italian',
    'ja': 'Japanese',
    'kk': 'Kazakh',
    'ko': 'Korean',
    'ky': 'Kyrgyz',
    'lv': 'Latvian',
    'lt': 'Lithuanian',
    'mk': 'Macedonian',
    'ml': 'Malay',
    'mn': 'Mongolian',
    'ne': 'Nepali',
    'nb': 'Norwegian-Bokmal',
    'nn': 'Norwegian-Nynorsk',
    'ps': 'Pashto',
    'fa': 'Persian',
    'pl': 'Polish',
    'pt-br': 'Portuguese-Brasil',
    'pt': 'Portuguese',
    'ro': 'Romanian',
    'ru': 'Russian',
    'sr': 'Serbian',
    'si': 'Sinhala',
    'sk': 'Slovak',
    'sl': 'Slovenian',
    'es': 'Spanish',
    'sw': 'Swahili',
    'sv': 'Swedish',
    'ta': 'Tamil',
    'te': 'telugu',
    'th': 'Thai',
    'tr': 'Turkish',
    'uk': 'Ukrainian',
    'ur': 'Urdu',
    'uz': 'Uzbek',
    'vi': 'Vietnamese',
    'cy': 'Welsh',
}


class I18nPlugin(TablePlugin):
    """Internationalization plugin."""

    def __init__(self, lang_code=None):
        """
        :param str lang_code: Language code (usually two-letter). E.g.: ru, de
        """
        self.lang = lang_code

    def contribute_to_config(self, config, table):
        super().contribute_to_config(config, table)

        code = self.lang

        if code is None:
            # Try auto.
            code = get_language()

        name = LANGS.get(code)

        if name:
            config['language'] = {'url': table.url_plugins + 'i18n/' + name + '.json'}



from django import template
from django.template import TemplateSyntaxError
from django.utils.safestring import SafeText

from ..toolbox.utils import dump_js

if False:  # pragma: nocover
    from ..toolbox import Table


register = template.Library()


def get_asset(attrname, tables):
    assets = {}

    for table in tables:  # type: Table
        for asset in getattr(table, attrname)():
            assets[asset] = ''  # Making a unique list.

    return SafeText(' '.join(assets.keys()))


def sitetable_tag(func):

    func_name = func.__name__

    def sitetable_tag_(table):

        if isinstance(table, str):
            raise TemplateSyntaxError(
                '`%s` template tag expects sitetable object. `%s` is given instead' % (func_name, type(table)))

        return SafeText(func(table))

    return register.simple_tag(sitetable_tag_, name=func_name)


@register.simple_tag()
def sitetables_js(*args):
    """

    :param list[Table] args:
    :rtype: str
    """
    return get_asset('get_assets_js', args)


@register.simple_tag()
def sitetables_css(*args):
    """

    :param list[Table] args:
    :rtype: str
    """
    return get_asset('get_assets_css', args)


@sitetable_tag
def sitetable_config(table):
    """

    :param Table table:
    :rtype: str
    """
    return dump_js(table.get_config())


@sitetable_tag
def sitetable_head(table):
    """

    :param Table table:
    :rtype: str
    """
    if not table.source.options.get('init_dom'):
        return ''

    return '<tr><th>%s</th></tr>' % '</th><th>'.join('%s' % column.title for column in table.get_columns().values())


@sitetable_tag
def sitetable_rows(table):
    """

    :param Table table:
    :rtype: str
    """
    if not table.source.options.get('init_dom'):
        return ''

    cols = list(table.get_columns().keys())
    return ''.join('<tr>%s</tr>' % ''.join('<td>%s</td>' % row[col] for col in cols) for row in table.get_rows())



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SitetablesConfig(AppConfig):
    """Application configuration."""

    name = 'sitetables'
    verbose_name = _('Sitetables')



import pytest
from django.template import TemplateSyntaxError

from sitetables.plugins.style.bootstrap4 import Bootstrap4Plugin
from sitetables.toolbox import Table


@pytest.fixture()
def render(template_render_tag, template_context):

    def render_(table, tag):
        return template_render_tag('sitetables', tag + ' table', context=template_context({'table': table}))

    return render_


def test_assets(template_render_tag, template_context, source_listdics):

    table1 = Table(
        source=source_listdics,
        plugins=[Bootstrap4Plugin()]
    )

    table2 = Table(
        source=source_listdics,
        plugins=[Bootstrap4Plugin()]
    )

    context = template_context({'table1': table1, 'table2': table2})

    rendered_js = template_render_tag('sitetables', 'sitetables_js table1 table2', context=context)

    assert 'jquery.dataTables' in rendered_js
    assert 'bootstrap4' in rendered_js
    assert str(rendered_js).count('script') == 4

    rendered_css = template_render_tag('sitetables', 'sitetables_css table1 table2', context=context)

    assert 'jquery.dataTables' in rendered_css
    assert 'bootstrap4' in rendered_css
    assert str(rendered_css).count('link') == 2


def test_head(render, source_listdics):

    assert render(Table(source_listdics), tag='sitetable_head') == ''  # Because of no DOM init option.

    rendered = render(Table(
        source={
            'source': source_listdics,
            'options': {'init_dom': True},
        }
    ), tag='sitetable_head')

    assert '<th>one</th>' in rendered


def test_rows(render, source_listdics):

    assert render(Table(source_listdics), tag='sitetable_rows') == ''  # Because of no DOM init option.

    rendered = render(Table(
        source={
            'source': source_listdics,
            'options': {'init_dom': True},
        }
    ), tag='sitetable_rows')

    assert '<td>a</td>' in rendered


def test_sitetable_tag(template_render_tag):

    with pytest.raises(TemplateSyntaxError):
        template_render_tag('sitetables', 'sitetable_config table')


def test_config(render, source_listdics):

    rendered = render(Table(source_listdics), tag='sitetable_config')

    assert '"columns":' in rendered
    assert '"data":' in rendered
    assert '"name": "one"' in rendered
    assert '"one": "a"' in rendered
    assert "function" in rendered
    assert '"function' not in rendered



from django.db import models


class Article(models.Model):

    title = models.CharField(max_length=20)
    hidden = models.BooleanField(default=False)

    def __str__(self):
        return self.title



from ..base import TablePlugin


class SemanticPlugin(TablePlugin):

    assets = {
        'js': [
            'dataTables.semanticui.min.js',
        ],
        'css': [
            'dataTables.semanticui.min.css',
        ],
    }



from collections import OrderedDict

import pytest

from pytest_djangoapp import configure_djangoapp_plugin


pytest_plugins = configure_djangoapp_plugin()


@pytest.fixture()
def source_listdics():
    out = [
        OrderedDict([
            ('one', 'a'),
            ('two', 'b'),
            ('three', 'c'),
        ]),
        OrderedDict([
            ('one', 'd'),
            ('two', 'e'),
            ('three', 'f'),
        ]),
    ]
    return out


@pytest.fixture()
def setup_articles():
    from sitetables.tests.testapp.models import Article

    def setup_(count=5):

        for idx in range(count):
            article = Article(title='my %s' % idx)
            article.save()

    return setup_



from django.db.models import Model, QuerySet

from .sources import ModelSource, ListDictsSource
from ..exceptions import SiteTablesException
from ..settings import URL_CDN_BASE

if False:  # pragma: nocover
    from ..plugins.base import TablePlugin


class Table:
    """Represents a sitetable."""

    version = '1.10.19'

    assets = {
        'js': [
            'jquery.dataTables.min.js',
        ],
        'css': [
            'jquery.dataTables.min.css',
        ],
    }

    def __init__(self, source=None, plugins=None, name=None):
        """
        :param dict|list|Model|QuerySet source: Table data source.

        :param list[TablePlugin] plugins: A list of plugins.

        :param str name: Table name (alias) to differentiate among
            different tables. If not set, default will be used.

        """
        self.source = None
        self.set_source(source)
        self.plugins = plugins or []
        self.url_base = URL_CDN_BASE
        self.name = name or 'table'

        # todo options argument maybe as a class?

    def set_source(self, source):
        """Sets a data source for this table.

        :param dict|list|Model|QuerySet source: Table data source.

        """
        if source is None:
            return None

        params = {}
        if isinstance(source, dict):
            params = source
            source = source['source']

        if isinstance(source, list):
            source_cls = ListDictsSource

        elif isinstance(source, QuerySet) or (type(source) == type(Model)):
            source_cls = ModelSource

        else:
            raise SiteTablesException('Unsupported data source type: %s' % type(source))

        self.source = source_cls.spawn(source, params)

    def get_config(self):
        """Returns a table configuration dictionary including params from plugins.

        :rtype: dict
        """
        source = self.source

        config = {}

        source.contribute_to_config(config, self)

        for plugin in self.plugins:
            plugin.contribute_to_config(config, table=self)

        return config

    def get_columns(self):
        """Returns columns dictionary from data source.

        :rtype: OrderedDict
        """
        return self.source.columns

    def get_rows(self):
        """Returns table rows from data source.

        :rtype: list[dict]

        """
        return self.source.rows

    @property
    def url_version(self):
        """Base URL for current DataTables version.

        :rtype: str
        """
        return self.url_base + self.version + '/'

    @property
    def url_plugins(self):
        """Base URL for plugins for current DataTables version.

        :rtype: str
        """
        return self.url_base + 'plug-ins/' + self.version + '/'

    def get_assets(self, typename, enclosure):
        """Returns a list of assets inclusion directives with their URLs.

        :param str typename: Assets type name. E.g.: js, css.
        :param str enclosure: Pattern to place URL into.

        :rtype: list
        """
        # todo checksum origin support

        assets = []

        url_asset_base = self.url_version + typename + '/'

        def contribute_assets(src):
            for asset in src.assets.get(typename, []):
                assets.append(enclosure % {'url': url_asset_base + asset})

        contribute_assets(self)

        for plugin in self.plugins:
            contribute_assets(plugin)

        return assets

    def get_assets_js(self):
        """Returns a list of HTML directives to include JS relevant for table and plugins.

        :rtype: list
        """
        return self.get_assets('js', '<script src="%(url)s"></script>')

    def get_assets_css(self):
        """Returns a list of HTML directives to include CSS relevant for table and plugins.

        :rtype: list
        """
        return self.get_assets('css', '<link rel="stylesheet" href="%(url)s">')



from ..base import TablePlugin


class Bootstrap3Plugin(TablePlugin):

    assets = {
        'js': [
            'dataTables.bootstrap.min.js',
        ],
        'css': [
            'dataTables.bootstrap.min.css',
        ],
    }



from django.utils.translation import activate, deactivate_all

from sitetables.plugins.i18n import I18nPlugin
from sitetables.plugins.style.bootstrap3 import Bootstrap3Plugin
from sitetables.plugins.style.bootstrap4 import Bootstrap4Plugin
from sitetables.plugins.style.foundation import FoundationPlugin
from sitetables.plugins.style.semantic import SemanticPlugin
from sitetables.plugins.style.jqueryui import JqueryUiPlugin
from sitetables.toolbox import Table


def test_jqueryui(source_listdics):

    table = Table(
        source=source_listdics,
        plugins=[
            JqueryUiPlugin(),
        ]
    )

    assert len(table.get_assets_css()) == 2

    js = table.get_assets_js()
    assert len(js) == 2
    assert 'jqueryui.' in js[1]


def test_semantic(source_listdics):

    table = Table(
        source=source_listdics,
        plugins=[
            SemanticPlugin(),
        ]
    )

    assert len(table.get_assets_css()) == 2

    js = table.get_assets_js()
    assert len(js) == 2
    assert 'semanticui.' in js[1]


def test_foundation(source_listdics):

    table = Table(
        source=source_listdics,
        plugins=[
            FoundationPlugin(),
        ]
    )

    assert len(table.get_assets_css()) == 2

    js = table.get_assets_js()
    assert len(js) == 2
    assert 'foundation.' in js[1]


def test_bootstrap3(source_listdics):

    table = Table(
        source=source_listdics,
        plugins=[
            Bootstrap3Plugin(),
        ]
    )

    assert len(table.get_assets_css()) == 2

    js = table.get_assets_js()
    assert len(js) == 2
    assert 'bootstrap.' in js[1]


def test_bootstrap4(source_listdics):

    table = Table(
        source=source_listdics,
        plugins=[
            Bootstrap4Plugin(),
        ]
    )

    assert len(table.get_assets_css()) == 2

    js = table.get_assets_js()
    assert len(js) == 2
    assert 'bootstrap4.' in js[1]


def test_i18n(source_listdics, settings):

    table = Table(source=source_listdics, plugins=[I18nPlugin()])

    assert len(table.get_assets_css()) == 1
    assert len(table.get_assets_js()) == 1

    config = table.get_config()
    assert 'language' not in config  # Unable to autodetect lang

    activate('ru')

    # Autodetect.
    config = table.get_config()
    assert 'Russian' in config['language']['url']

    # Manual.
    config = Table(source=source_listdics, plugins=[I18nPlugin('de')]).get_config()
    assert 'German' in config['language']['url']

    deactivate_all()




VERSION = (0, 1, 0)
"""Application version number tuple."""

VERSION_STR = '.'.join(map(str, VERSION))
"""Application version number string."""


default_app_config = 'sitetables.config.SitetablesConfig'



import json


class JsEncoder(json.JSONEncoder):
    """Used by dump_js, applies special markup to callables to make them js functions."""

    def default(self, obj):

        if callable(obj):
            return '<--%s-->' % obj()

        return super().default(obj)


def dump_js(obj):
    """Dumps dictionary to JS object.

    :param dict obj:
    :rtype: str
    """
    return json.dumps(obj, cls=JsEncoder).replace('"<--', '').replace('-->"', '')


def contribute_to_dict(tuples, to):
    """Updates target dict using data from tuples.

    :param list[tuple] tuples: List of (key, value) tuples.
    :param dict to: Target dict.

    """
    for name, value in tuples:
        if value is not None:
            to[name] = value



import pytest

from sitetables.exceptions import SiteTablesException
from sitetables.tests.testapp.models import Article
from sitetables.toolbox import Table


def test_unsupported_source():

    with pytest.raises(SiteTablesException):
        Table('unsupported')


def test_source_deferred(source_listdics):

    table = Table()
    table.set_source(source_listdics)

    assert isinstance(table.get_rows(), list)


def test_source_listdicts(source_listdics):

    table = Table(
        source=source_listdics
    )
    rows = table.get_rows()

    assert len(rows) == 2
    assert rows[0]['one'] == 'a'
    assert rows[1]['two'] == 'e'
    assert table.source.row_id == 'one'


def test_source_model(setup_articles):

    source = Article

    table = Table(source)
    assert not table.get_rows()  # No articles in DB.

    setup_articles()

    table = Table(source)
    rows = list(table.get_rows())

    assert len(rows) == 5
    assert table.source.row_id == 'id'


def test_source_query_set(setup_articles):

    source = Article.objects.filter(hidden=False)

    assert not Table(source).get_rows()  # No articles in DB.

    setup_articles()

    assert len(Table(source).get_rows()) == 5

    article = Article(title='hidden')
    article.save()

    assert len(Table(source).get_rows()) == 6

    article.hidden = True
    article.save()

    table = Table(source)
    assert len(table.get_rows()) == 5  # Still 5 due to filtering
    assert table.source.row_id == 'id'




if False:  # pragma: nocover
    from ..sitetable import Table


class TablePlugin:
    """Base class for plugins."""

    assets = {}
    """For additional plugin-related assets (js, css).

    Example:    
        assets = {
            'js': [
                'some.js',
            ],
            'css': [
                'some.css',
            ],
        }
    
    """

    def contribute_to_config(self, config, table):
        """Allows updating base table configuration dictionary with
        values from plugin specific.

        :param dict config:
        :param Table table:

        """
